#ifndef lint
static char sccs_id[] = "@(#) transform.c 1.2@(#)";
#endif /* lint */

/***
  transform.c - zooms & rotations for xtiff
***/
#include <stdio.h>
#include <X11/Intrinsic.h>	/* for typedefs */

#ifdef __STDC__
#define PP(args)	args
#else
#define PP(args)	()
#endif	/*__STDC__*/

#ifndef MAX
#define MAX(a, b) ( (a) > (b)  ?  (a)  : (b) )
#endif	/* MAX */
#ifndef MIN
#define MIN(a, b) ( (a) < (b)  ?  (a)  : (b) )
#endif	/* MIN */

int transformmimage PP((int status));
int zoomimage PP((void));
int rotateimage PP((void));
void rotate90 PP((int width, int height, int padding));
void rotate180 PP((int width, int height));
void rotate270 PP((int width, int height, int padding));
void setzoomfactor PP((int newzoomfactor));
int getzoomfactor PP((void));
void setrotationfactor PP((int newrotationfactor));
int getrotationfactor PP((void));
u_long getwidth PP((void));
u_long getheight PP((void));
u_long getrealwidth PP((void));
u_long getrealheight PP((void));
int getwidthpadding PP((void));
void seteffectivewidth PP((u_long length));
void seteffectiveheight PP((u_long length));

/*
  status flags from CreateXImage
*/
enum { INITSTATE = 0, PAGECHANGE = 1, ZOOMCHANGE = 2, ROTATECHANGE = 3 };

/*
 local shared data
*/
static int	zoomfactor = 1;
static int	rotateto = 0;
static int	currentrotation = 0;
static u_long	effectiveheight;
static u_long   effectivewidth;
 
/*
  shared data (about the image)
*/
extern u_char	*imageMemory,
		*zoomMemory;
extern u_long	tfImageHeight,
		tfImageWidth;
extern int	isize;
extern int	imagebitsize;


/*
  X data
*/
extern int	xImageDepth;

/*
  TIFF image data
*/
extern u_long	tfImageWidth, tfImageHeight;
extern u_short	tfBytesPerRow;

/*
  getbit(), setbit() for bit vectors, considered to be arrays of bits indexed
  from left to right, so bit 0 of bitvect bv is actually bit 7 (1<<7 = 0x80),
  the MSB, of bv[0].  there is a vague attempt to speed these up, with >> 3
  instead of / 8 and & 7 instead of % 8
*/
#define getbit(bv, bit) (0 != ( bv[ bit >> 3 ] & (0x80 >> (bit & 7)) ))
#define setbit(bv, bit)	bv[ bit >> 3] |= ( 0x80 >> (bit & 7) )

int
transformimage(status)
int status;
{
  if( PAGECHANGE == status )
	  /*
	    tell rotateimage() it's rotating from 0 to rotateto
	  */
	    currentrotation = 0;

  if( PAGECHANGE == status || ROTATECHANGE == status )
  	if( !rotateimage() )
		return 0;

	  if( !zoomimage() )
		return 0;

  return 1;

} /* transformimage */

void
rotate90(width, height)
u_long	width,
	height;
{
 register int	imgel,
		zmel,
		ibit,
		lastimgel;
 int		imagebytesize;
 int		incwidth;
 int 		pixels = width * height;

 switch(imagebitsize) {

	case 1: imgel = lastimgel = width * height - width;
		for(zmel=0; zmel<pixels; zmel++) {

			/*
			  testing for & not transforming a blank line
			  has shown no great speed advantage
			*/
			if( getbit(imageMemory, imgel) )
				setbit(zoomMemory, zmel);

			if( 0 > (imgel -= width) ) 
				imgel = ++lastimgel;

		} /* for zmel */

		break;

	case 8: imgel = lastimgel = width * height - width;
		for(zmel=0; zmel<width*height; zmel++) {

			zoomMemory[zmel] = imageMemory[imgel];

			if( 0 > (imgel -= width) )
				imgel = ++lastimgel;

		} /* for zmel */

		break;

dividesby8:	
	case 16:
	case 24:
	case 32:imagebytesize = imagebitsize / 8;
		imgel = lastimgel = width * height - width;
		/*
		  speed things up a little in the loop
		*/
		  incwidth = width * imagebytesize;

		for(zmel=0; zmel<width*height*imagebytesize; 
							zmel+=imagebytesize) {

			(void)memcpy(zoomMemory + zmel, imageMemory + imgel,
								imagebytesize);

			if( 0 > (imgel -= incwidth) )
				imgel = (lastimgel += imagebytesize);

		} /* for zmel */

		break;

	default: /* arbitrary groups of bits ... */

		/*
		  ... but not for this routine if arbitrary groups of bytes
		*/
		  if( !(imagebitsize % 8) )
			goto dividesby8;

		/*
		  speed up the loop a little
		*/
		  incwidth = width * imagebitsize;

	 	imgel = lastimgel = (width * height - width) * imagebitsize;
		for(zmel=0; zmel<width*height*imagebitsize; zmel++) {

			for(ibit=imgel; ibit<imgel+imagebitsize; ibit++) {

				if( getbit(imageMemory, ibit) )
					setbit(zoomMemory, zmel);

				++zmel;

			} /* for ibit */

			if( 0 > (imgel -= incwidth) )
				imgel = (lastimgel += imagebitsize);

		} /* for zmel */

 } /* switch imagebitsize */

} /* rotate90 */

void
rotate180(width, height)
u_long	width,
	height;
{
 register int	imgel,
		zmel,
		ibit;
 int		imagebytesize,
		pixels = width * height;

 switch(imagebitsize) {

	case 1: imgel =  width * height - 1;
		for(zmel=0; zmel<pixels; zmel++) {

			if( getbit(imageMemory, imgel) )
				setbit(zoomMemory, zmel);

			--imgel;

		} /* for zmel */

		break;

	case 8: imgel = width * height - 1;
		for(zmel=0; zmel<width*height; zmel++) 
			zoomMemory[zmel] = imageMemory[imgel--];

		break;

dividesby8:	
	case 16:
	case 24:
	case 32:imagebytesize = imagebitsize / 8;
		imgel = width * height - imagebytesize;

		for(zmel=0; zmel<width*height*imagebytesize; 
							zmel+=imagebytesize) {

			(void)memcpy(zoomMemory + zmel, imageMemory + imgel,
								imagebytesize);

			imgel -= imagebytesize;

		} /* for zmel */

		break;

	default: /* arbitrary groups of bits ... */

		/*
		  ... but not for this routine if arbitrary groups of bytes
		*/
		  if( !(imagebitsize % 8) )
			goto dividesby8;

	 	imgel = (width * height - 1) * imagebitsize;
		for(zmel=0; zmel<width*height*imagebitsize; zmel++) {

			for(ibit=imgel; ibit<imgel+imagebitsize; ibit++) {

				if( getbit(imageMemory, ibit) )
					setbit(zoomMemory, zmel);

				++zmel;

			} /* for ibit */

			imgel -= imagebitsize;

		} /* for zmel */

 } /* switch imagebitsize */

} /* rotate180 */


void
rotate270(width, height)
u_long	width,
	height;
{
 register int	imgel,
		zmel,
		ibit,
		lastimgel;
 int		imagebytesize;
 int		incwidth,
		incpoint,
		pixels = width * height;

 switch(imagebitsize) {

	case 1: imgel = lastimgel = width - 1;
		for(zmel=0; zmel<pixels; zmel++) {

			if( getbit(imageMemory, imgel + zmel) )
				setbit(zoomMemory, zmel);

			imgel += width - 1;

			if( !((zmel + 1) % height) )
				imgel = (lastimgel -= (height + 1));

		} /* for zmel */

		
		break;

	case 8: imgel = lastimgel = width - 1;
		for(zmel=0; zmel<width*height; zmel++) {

			zoomMemory[zmel] = imageMemory[imgel];

			imgel += width - 1;

			if( !((zmel + 1) % height) ){
				imgel = lastimgel - (height + 1);
				lastimgel -= height + 1;
			} /* if zmel */

		} /* for zmel */
		break;

dividesby8:	
	case 16:
	case 24:
	case 32:imagebytesize = imagebitsize / 8;
		imgel = lastimgel = (width - 1) * imagebytesize;

		/*
		  speed things up a little in the loop
		*/
		  incwidth = (width - 1) * imagebytesize;
		  incpoint = height * imagebytesize;

		for(zmel=0; zmel<width*height*imagebytesize; 
							zmel+=imagebytesize) {

			(void)memcpy(zoomMemory + zmel, imageMemory + imgel,
								imagebytesize);

			imgel += incwidth;

			if( !((zmel + 1) % incpoint) ){
				imgel =    lastimgel 
					- (height + 1) * imagebytesize;
				lastimgel -= (height + 1) * imagebytesize;
			} /* if zmel */

		} /* for zmel */

		break;

	default: /* arbitrary groups of bits ... */

		/*
		  ... but not for this routine if arbitrary groups of bytes
		*/
		  if( !(imagebitsize % 8) )
			goto dividesby8;

		/*
		  speed up the loop a little
		*/
		  incwidth = width * imagebitsize;

	 	imgel = lastimgel = (width - 1) * imagebitsize;
		for(zmel=0; zmel<width*height*imagebitsize; zmel++) {

			for(ibit=imgel; ibit<imgel+imagebitsize; ibit++) {

				if( getbit(imageMemory, ibit) )
					setbit(zoomMemory, zmel);

				++zmel;

			} /* for ibit */

			imgel += (width - 1) * imagebitsize;

			if( !((imgel + 1) % (height * imagebitsize)) ) {
	
				imgel = lastimgel - (height - 1) * imagebitsize;
				lastimgel -= (height - 1) * imagebitsize;

			} /* if ! imgel */

		} /* for zmel */

 } /* switch imagebitsize */

} /* rotate270 */

int
rotateimage()
{
 int	rotationdelta;
 u_long	previousheight = effectiveheight;
 u_long	previouswidth = effectivewidth;	

 /*
   calculate the width & height of the image we're rotating
 */
   if( 90 == currentrotation || 270 == currentrotation ) {
	previousheight = effectivewidth;
	previouswidth = effectiveheight;
   } /* if 90 */

 if( rotateto == currentrotation )
	return 1;

 /*
   image may already be rotated, so work out rotation factor to rotate it to
   the angle the user wants
 */
   rotationdelta = rotateto - currentrotation;

   if( 0 > rotationdelta )
	rotationdelta += 360; 

   zoomMemory = (u_char *) malloc(isize);

   if( NULL == zoomMemory ) {
	(void)fprintf(stderr, "xtiff: not enough memory for rotate\n");
	rotateto = currentrotation;
	return 0;
   } /* if NULL */

   (void)memset(zoomMemory, 0, isize);

 switch( rotationdelta ) {

	case 90:
		rotate90(previouswidth, previousheight);
		break;
	case 180:
		rotate180(previouswidth, previousheight);
		break;
	case 270:
		rotate270(previouswidth, previousheight);
		break;
	default:	
		(void)fprintf(stderr, "xtiff:false rotationdelta %d", 
								rotationdelta);
 
 } /* switch rotationdelta */

 /*
   copy that lot back into the image memory
 */
   (void)memcpy(imageMemory, zoomMemory, isize);

 free(zoomMemory);
 zoomMemory = NULL;

 currentrotation = rotateto;

 return 1;	

} /* rotateimage */

int
zoomimage()
{
   register int	i,
		j,
		zbit = 7,
		ibit;
   int		imagebytesize,
		memsize;
   u_char	*zmp,
		*imp = imageMemory;
   u_long	width = getrealwidth(),
		height = getrealheight();
   u_long	bytesperrow = getrealwidth() / 8;

   if( NULL != zoomMemory ) 
	free(zoomMemory);

   zoomMemory = NULL;

   if( 1 == zoomfactor ) {

	zoomMemory = (u_char *)malloc(isize);
        if( NULL == zoomMemory ) {
		(void)fprintf(stderr, "xtiff: not enough memory for zoom\n");
		return 0;
	} /* if NULL */

	(void)memcpy(zoomMemory, imageMemory, isize);
	return 1;

   } /* if 1 */

  
   switch( imagebitsize ) {

	case 1:	memsize = ((getwidth() + 7) * (getheight() + 7)) / 8;

		zoomMemory = (u_char *) malloc(memsize);
 
		zmp = zoomMemory;

		if( NULL == zmp ) {
			(void)fprintf(stderr, 
					"xtiff: not enough memory for zoom\n");
			return 0;
		} /* if NULL */

		(void)memset(zmp, 0, memsize); 

		for(i=0; i<height; i+=zoomfactor) {

			for(j=0; j<width; j+=zoomfactor) {
			
				/*
				  set this bit of zoomMemory to resampled
				  bit of imageMemory
				*/		
				  if( getbit(imp, j) )
				  	*zmp |= (1 << zbit);

				/*
				  keep looping through bit numbers
				*/	
				  if( 0 > --zbit ) {
					++zmp;
					zbit = 7;
				  } /* if !zbit */

			} /* for j */

			/*
			  if we're not at the end of the last byte of the zoom
			  scan line move to next byte, X wants images like this
			*/
			  if( 7 != zbit ) {
				if( 6 != zbit )		/* KLUDGE */
					++zmp;
				zbit = 7;
			  } /* if 7 */

			imp += zoomfactor * bytesperrow;

		} /* for i */

		return 1;

	case 8:	memsize = getwidth() * getheight();

		zoomMemory = (u_char *) malloc(memsize);

		zmp = zoomMemory;

		if( NULL == zmp )
			return 0;

		(void)memset(zmp, 0, memsize);

		for(i=0; i<height; i+=zoomfactor) {

			for(j=0; j<width; j+=zoomfactor)
				*zmp++ = imp[j];

			imp += zoomfactor * width;

		} /* for i */

		return 1;

dividesby8:
	case 16:
	case 24:
	case 32:imagebytesize = imagebitsize / 8;
		memsize = getwidth() * getheight() * imagebytesize;
		zoomMemory = (u_char *) malloc(memsize);

		zmp = zoomMemory;

		if( NULL == zmp )
			return 0;

		(void)memset(zmp, 0, memsize); 

		for(i=0; i<height; i+=zoomfactor) {

			for(j=0; j<width; j+=zoomfactor)
				(void)memcpy(zmp, imp + j * imagebytesize,
					imagebytesize);

			imp += zoomfactor * width * imagebytesize;

		} /* for i */

		return 1;

	default: /* some arbitrary number of bits ... */
		 /* 
		   ... but let's just check if it's groups of bytes, a
		   96-bit display or something ...
		 */
		   if( !(imagebitsize % 8) )
			goto dividesby8;

		 memsize = (getwidth() * getheight() * imagebitsize) / 8
			  + getheight();

		 zoomMemory = (u_char *) malloc(memsize); 

		 zmp = zoomMemory;

		 if( NULL == zmp )
			return 0;

		(void)memset(zmp, 0, memsize);

		for(i=0; i<height; i+=zoomfactor) {

			for(j=0; j<width*imagebitsize; 
				j+=zoomfactor*imagebitsize) 
			
				for(ibit=j; ibit<j+imagebitsize; ibit++) {

					  *zmp |= (getbit(imp, ibit)<<zbit);

					/*
					  keep looping through bit numbers
					*/	
					  if( 0 > --zbit ) {
						++zmp;
						zbit = 7;
					  } /* if !zbit */

				} /* for ibit */

			/*
			  if we're not at the end of the last byte of the zoom
			  scan line move to next byte, X wants images like this
			*/
			  if( 7 != zbit ) {
				++zmp;
				zbit = 7;
			  } /* if 7 */

			imp += zoomfactor * bytesperrow;

		} /* for i */

		return 1;

    } /* switch xImageDepth */

} /* zoomimage */


void
setzoomfactor(newzoomfactor)
int newzoomfactor;
{
 if( 0 < newzoomfactor )
	zoomfactor = newzoomfactor;

} /* setzoomfactor */


int 
getzoomfactor()
{
 return zoomfactor;
} /* getzoomfactor */

void
setrotationfactor(newrotationfactor)
int newrotationfactor;
{
 if(      0 != newrotationfactor
     &&  90 != newrotationfactor 
     && 180 != newrotationfactor 
     && 270 != newrotationfactor )
	return;

 rotateto = newrotationfactor;
} /* setrotationfactor */

int
getrotationfactor()
{
 return rotateto;
} /* getrotationfactor */

u_long
getrealwidth()
{
 return (    90 == rotateto 
	 || 270 == rotateto ) ? effectiveheight : effectivewidth;
} /* getrealwidth */

u_long
getrealheight()
{
 return (    90 == rotateto 
         || 270 == rotateto ) ? effectivewidth : effectiveheight;
} /* getrealheight */

u_long
getwidth()
{
 return getrealwidth() / zoomfactor;
} /* getwidth */

u_long
getheight()
{
 return getrealheight() / zoomfactor;
} /* getheight */


void
seteffectivewidth(length)
u_long length;
{
 effectivewidth = length;
} /* seteffectivewidth */

void
seteffectiveheight(length)
u_long length;
{
 effectiveheight = length;
} /* seteffectiveheight */

/*** transform.c ***/

